package eu.etaxonomy.taxeditor.prototype2.view;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.util.List;
import java.util.Scanner;

import org.eclipse.core.databinding.Binding;
import org.eclipse.core.databinding.DataBindingContext;
import org.eclipse.core.databinding.beans.BeansObservables;
import org.eclipse.core.databinding.observable.value.IObservableValue;
import org.eclipse.core.resources.IMarker;
import org.eclipse.core.runtime.Assert;
import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.Document;
import org.eclipse.jface.text.IDocument;
import org.eclipse.jface.text.IDocumentPartitioner;
import org.eclipse.jface.text.ITypedRegion;
import org.eclipse.jface.text.Position;
import org.eclipse.jface.text.TypedRegion;
import org.eclipse.jface.text.rules.DefaultPartitioner;
import org.eclipse.jface.text.rules.FastPartitioner;
import org.eclipse.jface.text.rules.IPartitionTokenScanner;
import org.eclipse.jface.text.source.Annotation;
import org.eclipse.jface.text.source.AnnotationModel;
import org.eclipse.jface.text.source.AnnotationPainter;
import org.eclipse.jface.text.source.IAnnotationAccess;
import org.eclipse.jface.text.source.IAnnotationAccessExtension;
import org.eclipse.jface.text.source.ImageUtilities;
import org.eclipse.jface.text.source.SourceViewer;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.StyleRange;
import org.eclipse.swt.custom.StyledText;
import org.eclipse.swt.events.FocusEvent;
import org.eclipse.swt.events.FocusListener;
import org.eclipse.swt.events.KeyAdapter;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.RGB;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.widgets.Canvas;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;

import com.swtdesigner.SWTResourceManager;

import eu.etaxonomy.taxeditor.prototype2.view.nameviewersupport.AnnotationMarkerAccess;
import eu.etaxonomy.taxeditor.prototype2.view.nameviewersupport.ErrorAnnotation;
import eu.etaxonomy.taxeditor.prototype2.view.nameviewersupport.NamePartitionScanner_;

/**
 * Creates zero-length fieldPartitions for Genus, Specific Epithet, etc., then fills them
 *  as data is added. 
 *  
 * Partition data are bound to their respective model properties.
 * 
 * @author p.ciardelli
 *
 */
/**
 * @author p.ciardelli
 *
 */
public class NameViewer_ extends SourceViewer {
	
	final Document document;
	final IDocumentPartitioner partitioner;
	private TaxonName taxonName;
	private DataBindingContext bindingContext;
	protected Composite parent;
	
	private BoundPartition[] fieldPartitions;
	
	StyledText styledText;

    // error identifiers, images and colors
    public static String ERROR_TYPE = "error.type";
    public static Image ERROR_IMAGE;
    public static final RGB ERROR_RGB = new RGB(255, 0, 0);
    
    // annotation model
    private AnnotationModel fAnnotationModel = new AnnotationModel();    
	
	NameViewer_(TaxonName taxonName, DataBindingContext bindingContext, Composite parent) {
		super(parent, null, SWT.WRAP | SWT.MULTI | SWT.RESIZE); // null -> no vertical ruler
		
		this.taxonName = taxonName;
		this.bindingContext = bindingContext;
		this.parent = parent;
				
        Assert.isNotNull(bindingContext,
            	"NameViewer_'s binding context must be explicitly set.");
		
		fieldPartitions = new BoundPartition[] {
				new BoundPartition("genus", SWT.ITALIC),
				new BoundPartition("specificEpithet", SWT.BOLD)
		};
		
		document = new Document(taxonName.getGenus() + " " + taxonName.getSpecificEpithet());
		
		partitioner = new NamePartitioner(new NamePartitionScanner_(),
				NamePartitionScanner_.PARTITION_TYPES);
		
		partitioner.connect(document);
		document.setDocumentPartitioner(partitioner);

		this.setDocument(document, fAnnotationModel); // put this back in nameeditorview so that it functions more like TextViewer
			
		styledText = this.getTextWidget();
//		styledText.setIndent(8);
		styledText.addFocusListener(new FocusListener() {

			public void focusGained(FocusEvent e) {
				((StyledText) e.widget).setBackground((Display.getCurrent().getSystemColor(SWT.COLOR_GRAY)));
				
			}

			public void focusLost(FocusEvent e) {
				((StyledText) e.widget).setBackground((Display.getCurrent().getSystemColor(SWT.COLOR_WHITE)));
			}
			
		});
		
		// annotations section
		IAnnotationAccess fAnnotationAccess = new AnnotationMarkerAccess();
		
		// to paint the annotations
        AnnotationPainter ap = new AnnotationPainter(this, fAnnotationAccess);
        ap.addAnnotationType(ERROR_TYPE);
        ap.setAnnotationTypeColor(ERROR_TYPE, new Color(Display.getDefault(), ERROR_RGB));

        // this will draw the squigglies under the text
        this.addPainter(ap);
        
        // add an annotation
        ErrorAnnotation errorAnnotation = new ErrorAnnotation(1, "Learn how to spell \"text!\"");

        // lets underline the word "texst"
        fAnnotationModel.addAnnotation(errorAnnotation, new Position(1, 5));
				
	};

	
	/**
	 * Separate out properties that are different for accepted and related names 
	 */
	private void setViewerLayoutExtras() {
		styledText.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false, 2, 1));
		styledText.setFont(SWTResourceManager.getFont("Georgia", 12, SWT.NONE));		
	}

	public void setBindingContext(DataBindingContext bindingContext) {
		this.bindingContext = bindingContext;
	}
	
	class NamePartitioner extends FastPartitioner {

		List<ITypedRegion> wordPartitions;

		public NamePartitioner(IPartitionTokenScanner scanner,
				String[] legalContentTypes) {
			super(scanner, legalContentTypes);
		}
		
		public ITypedRegion[] computePartitioning(int offset, int length, boolean includeZeroLengthPartitions) {

			int fieldIndex = 0;
			
//			System.out.print("Start re-partition at: " + offset + " -");
//			for (int i = 0; i < fieldPartitions.length; i++) {
//				System.out.print(" " + fieldPartitions[i].getOffset());
//			}
//			System.out.println();
//						
//			// find where the affected partitions begin
//			while (fieldPartitions[fieldIndex].getOffset() < offset) {
//				System.out.println(fieldPartitions[fieldIndex].getOffset() + " " + offset);
//				fieldIndex++;
//			}
			
//			ITypedRegion[] regions = super.computePartitioning(offset, document.getLength() - offset, 
//					includeZeroLengthPartitions);
			ITypedRegion[] regions = super.computePartitioning(0, document.getLength(), 
					includeZeroLengthPartitions);	
			
			for (int i = 0; i < regions.length; i++) {
				if (regions[i].getType() == NamePartitionScanner_.SINGLE_WORD) {
					// for now, assign first word to first fieldPartition, etc.
					fieldPartitions[fieldIndex].setRegion(regions[i]);
					fieldIndex++;
				}
			}	
			
			// put zero-length regions in remaining fieldPartitions 
			for (int i = fieldIndex; i < fieldPartitions.length; i++) {
				fieldPartitions[fieldIndex].setRegion(
						new TypedRegion(document.getLength(), 0, 
								NamePartitionScanner_.SINGLE_WORD));
			} 
			return regions;
		}
		
		
		public void connect(IDocument document, boolean delayInitialise)
		{
			super.connect(document, delayInitialise);
//			computePartitioning(0, document.getLength(), false);
		}		
	}
	
	class BoundPartition {
		private ITypedRegion region;
		private Binding binding;
		private String field; // the name field to track
		private String value; // the current value in this partition
		private int length;
		private int offset;
		private StyleRange style = new StyleRange();
				
		BoundPartition (String field) {
			this.field = field;
			this.length = 0;
			this.offset = 0;
			
			bindPartition();
		}

		BoundPartition (String field, int fontStyle) {
			this(field);
			style.fontStyle = fontStyle;
		}
		
		private void bindPartition() {
			IObservableValue observePartitionValue =
					BeansObservables.observeValue(this, "text");
			IObservableValue observeModelValue =
					BeansObservables.observeValue(taxonName, field);			
			this.binding = bindingContext.bindValue(
					observePartitionValue, observeModelValue, 
					null, null);	
		}
		
		public void setRegion (ITypedRegion region) {
			this.region = region;
			this.length = region.getLength();
			this.offset = region.getOffset();
						
			String oldValue = this.value; 
			try {
				this.value = document.get(this.offset, this.length);
			} catch (BadLocationException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			firePropertyChange("text",oldValue,this.value);

			this.style.start = this.offset;
			this.style.length = this.length;
			styledText.setStyleRange(this.style);
		}
		
		public int getLength () {
			return this.length;
		}
		
		public int getOffset () {
			return this.offset;
		}
		
		public void setText(String newText) {
			try {
				document.replace(region.getOffset(), region.getLength(), newText);
			} catch (BadLocationException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		
		public String getText() {
			return this.value;
		}
		
		protected final PropertyChangeSupport propertyChangeSupport = new PropertyChangeSupport(this);
		public void addPropertyChangeListener(PropertyChangeListener listener) {
			propertyChangeSupport.addPropertyChangeListener(listener);
		}

		public void addPropertyChangeListener(String propertyName,
				PropertyChangeListener listener) {
			propertyChangeSupport.addPropertyChangeListener(propertyName, listener);
		}

		public void removePropertyChangeListener(PropertyChangeListener listener) {
			propertyChangeSupport.removePropertyChangeListener(listener);
		}

		public void removePropertyChangeListener(String propertyName,
				PropertyChangeListener listener) {
			propertyChangeSupport.removePropertyChangeListener(propertyName,
					listener);
		}

		protected void firePropertyChange(String propertyName, Object oldValue,
				Object newValue) {
			propertyChangeSupport.firePropertyChange(propertyName, oldValue,
					newValue);
		}

//		public StyleRange getStyle() {
//			return style;
//		}
//
//		public void setStyle(StyleRange style) {
//			this.style = style;
//		}
	}
	
//    class AnnotationMarkerAccess implements IAnnotationAccess, IAnnotationAccessExtension {
//        public Object getType(Annotation annotation) {
//            return annotation.getType();
//        }
//
//        public boolean isMultiLine(Annotation annotation) {
//            return true;
//        }
//
//        public boolean isTemporary(Annotation annotation) {
//            return !annotation.isPersistent();
//        }
//
//        public String getTypeLabel(Annotation annotation) {
//            if (annotation instanceof ErrorAnnotation)
//                return "Errors";
//
//            return null;
//        }
//
//        public int getLayer(Annotation annotation) {
//            if (annotation instanceof ErrorAnnotation)
//                return ((ErrorAnnotation)annotation).getLayer();
//
//            return 0;
//        }
//
//        public void paint(Annotation annotation, GC gc, Canvas canvas, Rectangle bounds) {
//            ImageUtilities.drawImage(((ErrorAnnotation)annotation).getImage(), 
//            		gc, canvas, bounds, SWT.CENTER, SWT.TOP);
//        }
//
//        public boolean isPaintable(Annotation annotation) {
//            if (annotation instanceof ErrorAnnotation)
//                return ((ErrorAnnotation)annotation).getImage() != null;
//
//            return false;
//        }
//
//        public boolean isSubtype(Object annotationType, Object potentialSupertype) {
//            if (annotationType.equals(potentialSupertype))
//                return true;
//
//            return false;
//
//        }
//
//        public Object[] getSupertypes(Object annotationType) {
//            return new Object[0];
//        }
//    }
    
    // one error annotation
//    class ErrorAnnotation extends Annotation {
//        private IMarker marker;
//        private String text;
//        private int line;
//        private Position position;
//
//        public ErrorAnnotation(IMarker marker) {
//            this.marker = marker;
//        }
//
//        public ErrorAnnotation(int line, String text) {
//            super(ERROR_TYPE, true, null);
//            this.marker = null;
//            this.line = line;
//            this.text = text;
//        }
//
//        public IMarker getMarker() {
//            return marker;
//        }
//
//        public int getLine() {
//            return line;
//        }
//
//        public String getText() {
//            return text;
//        }
//
//        public Image getImage() {
//            return ERROR_IMAGE;
//        }
//
//        public int getLayer() {
//            return 3;
//        }
//
//        public String getType() {
//            return ERROR_TYPE;
//        }
//
//        public Position getPosition() {
//            return position;
//        }
//
//        public void setPosition(Position position) {
//            this.position = position;
//        }
//    }    
    
}
