/**
* Copyright (C) 2007 EDIT
* European Distributed Institute of Taxonomy 
* http://www.e-taxonomy.eu
* 
* The contents of this file are subject to the Mozilla Public License Version 1.1
* See LICENSE.TXT at the top of this package for the full license terms.
*/

package eu.etaxonomy.taxeditor;

import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.action.IMenuManager;
import org.eclipse.jface.action.MenuManager;
import org.eclipse.jface.action.Separator;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.actions.ActionFactory;
import org.eclipse.ui.actions.ActionFactory.IWorkbenchAction;
import org.eclipse.ui.application.ActionBarAdvisor;
import org.eclipse.ui.application.IActionBarConfigurer;
import org.eclipse.ui.operations.UndoActionHandler;

import eu.etaxonomy.cdm.io.common.ImportWrapper;
import eu.etaxonomy.taxeditor.actions.io.ExportAction;
import eu.etaxonomy.taxeditor.actions.io.ImportAction;
import eu.etaxonomy.taxeditor.actions.ui.OpenTaxonEditorAction;
import eu.etaxonomy.taxeditor.controller.GlobalController;

/**
 * An action bar advisor is responsible for creating, adding, and disposing of
 * the actions added to a workbench window. Each window will be populated with
 * new actions.
 * 
 * @author p.ciardelli
 * @created 02.06.2008
 * @version 1.0
 */
public class ApplicationActionBarAdvisor extends ActionBarAdvisor {
	private static final Logger logger = Logger
			.getLogger(ApplicationActionBarAdvisor.class);

	// Actions - important to allocate these only in makeActions, and then use
	// them in the fill methods. This ensures that the actions aren't recreated
	// when fillActionBars is called with FILL_PROXY.
	private IWorkbenchAction exitAction;
	private IWorkbenchAction saveAction;
	private IWorkbenchAction preferencesAction;
	private IWorkbenchAction undoAction;
	private IWorkbenchAction redoAction;
	
	private IAction newNameAction;

	private List<IAction> importActionList;

	private IAction exportJaxbAction;


	public ApplicationActionBarAdvisor(IActionBarConfigurer configurer) {
		super(configurer);
	}

	protected void makeActions(final IWorkbenchWindow window) {
		// Creates the actions and registers them.
		// Registering is needed to ensure that key bindings work.
		// The corresponding commands keybindings are defined in the plugin.xml
		// file.
		// Registering also provides automatic disposal of the actions when
		// the window is closed.

		GlobalController.setStatusLineManager(getActionBarConfigurer().getStatusLineManager());
		
		saveAction = ActionFactory.SAVE.create(window);
		register(saveAction);
		
		undoAction = ActionFactory.UNDO.create(window);
		register(undoAction);
		
		redoAction = ActionFactory.REDO.create(window);
		register(redoAction);
		
		exitAction = ActionFactory.QUIT.create(window);
		register(exitAction);
		
		newNameAction = new OpenTaxonEditorAction();
		register(newNameAction);
		
		preferencesAction = ActionFactory.PREFERENCES.create(window);
		
		makeImportActions();
		
		exportJaxbAction = new ExportAction(ExportAction.JAXB);
		register(exportJaxbAction);
	}

	private void makeImportActions() {
		
		importActionList = new ArrayList<IAction>();
		
		for (ImportWrapper wrapper : ImportWrapper.list()) {
			IAction importAction = new ImportAction(wrapper);
			register(importAction);
			importActionList.add(importAction);		
		}
	}

	protected void fillMenuBar(IMenuManager menuBar) {

		// Note: to hook into Eclipse File Menu, to use open File for instance,
		//  replace NULL with IWorkbenchActionConstants.M_FILE);
		MenuManager fileMenu = new MenuManager("&File",
				null);
		
		MenuManager editMenu = new MenuManager("&Edit", null);
		
		// Create submenu for imports
		MenuManager importMenu = new MenuManager("Import ...", null);
		
		// Create submenu for exports
		MenuManager exportMenu = new MenuManager("Export as ...", null);

		// Populate file menu
		menuBar.add(fileMenu);
		fileMenu.add(newNameAction);
		fileMenu.add(saveAction);
		fileMenu.add(new Separator());
		fileMenu.add(importMenu);
		fileMenu.add(exportMenu);
		fileMenu.add(new Separator());
		fileMenu.add(exitAction);
		
		menuBar.add(editMenu);
		editMenu.add(undoAction);
		editMenu.add(redoAction);
		
		// Populate submenu for imports
		for (IAction importAction : importActionList) {
			importMenu.add(importAction);
		}

		// Populate submenu for exports
		exportMenu.add(exportJaxbAction);
		
		// Populate preferences
		MenuManager preferencesMenu = new MenuManager("&Preferences",
				null);
		menuBar.add(preferencesMenu);
		preferencesMenu.add(preferencesAction);
	}
}