package eu.etaxonomy.taxeditor.editor;

import org.eclipse.core.databinding.DataBindingContext;
import org.eclipse.core.databinding.beans.BeansObservables;
import org.eclipse.jface.databinding.swt.SWTObservables;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.KeyListener;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Text;
import org.eclipse.ui.forms.IDetailsPage;
import org.eclipse.ui.forms.IManagedForm;
import org.eclipse.ui.forms.events.ExpansionAdapter;
import org.eclipse.ui.forms.events.ExpansionEvent;
import org.eclipse.ui.forms.widgets.ColumnLayout;
import org.eclipse.ui.forms.widgets.FormToolkit;
import org.eclipse.ui.forms.widgets.ScrolledForm;
import org.eclipse.ui.forms.widgets.Section;

/**
 * The abstract class for details, i.e. form-based editing.
 * 
 * Implementing classes should add section and input fields in
 *  selectionChanged(). Content elements are null before this.
 *  
 * Sections can be wiped clean to be re-built with disposeSectionContent(section),
 *  for instance, when a TaxonName changes its rank and a name consists of different
 *  parts. 
 * 
 * @author p.ciardelli
 *
 */
public abstract class EditorDetailsPage implements IDetailsPage {

	private FormToolkit toolkit;
	private ScrolledForm form;
	private IManagedForm mform;
	private Composite detailsComposite;
	private DataBindingContext bindingContext = new DataBindingContext();
	
	public void createContents(Composite parent) {	
		
		detailsComposite = parent;
		toolkit = mform.getToolkit();
		
		ColumnLayout layout = new ColumnLayout();
		layout.topMargin = 0;
		layout.bottomMargin = 5;
		layout.leftMargin = 10;
		layout.rightMargin = 10;
		layout.horizontalSpacing = 10;
		layout.verticalSpacing = 10;
		layout.maxNumColumns = 4;
		layout.minNumColumns = 1;				
		parent.setLayout(layout);	
	}

	protected Section createSection(String title) {
		final Section section = toolkit.createSection(detailsComposite, 
				Section.TWISTIE | Section.TITLE_BAR | Section.EXPANDED);
		section.marginWidth = 10;
		section.setText(title); //$NON-NLS-1$		
		
		createClient(section);
		
		section.addExpansionListener(new ExpansionAdapter() {
			public void expansionStateChanged(ExpansionEvent e) {
				form.reflow(false);
			}
		});
		
		return section;
	}
	
	private void createClient(Section section) {

		Composite client = toolkit.createComposite(section);
		GridLayout gridlayout = new GridLayout();
		gridlayout.marginWidth = gridlayout.marginHeight = 5;
		gridlayout.numColumns = 2;
		client.setLayout(gridlayout);
		section.setClient(client);
		
		toolkit.paintBordersFor(client);
	}
	
	protected Control createInputField(Section section, String label, 
			Object bindObject, String bindField) {
		
		Composite client = (Composite) section.getClient();
		toolkit.createLabel(client, label); //$NON-NLS-1$
		Text text = toolkit.createText(client, "", SWT.SINGLE );
		text.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false)); //$NON-NLS-1$
		text.setData(FormToolkit.TEXT_BORDER);
		
		text.addKeyListener(new KeyListener() {

			public void keyPressed(KeyEvent e) {
			}

			public void keyReleased(KeyEvent e) {
				showObject();
			}
			
		});
		
		if (!(bindObject == null)) {
			bindingContext.bindValue(
					SWTObservables.observeText(text, SWT.Modify),
					BeansObservables.observeValue(bindObject, bindField),
					null, null);
		}
		return text;
	}


	abstract public void showObject();
	
	protected void disposeSectionContent(Section section) {
		section.getClient().dispose();
		createClient(section);
	}
	
	public void initialize(IManagedForm mform) {
		this.mform = mform;
	}
}