/**
 * Copyright (C) 2007 EDIT
 * European Distributed Institute of Taxonomy 
 * http://www.e-taxonomy.eu
 * 
 * The contents of this file are subject to the Mozilla Public License Version 1.1
 * See LICENSE.TXT at the top of this package for the full license terms.
 */

package eu.etaxonomy.taxeditor.editor;

import org.apache.log4j.Logger;
import org.eclipse.core.runtime.Assert;
import org.eclipse.jface.text.IDocument;
import org.eclipse.jface.text.TextViewer;
import org.eclipse.jface.viewers.ViewerComparator;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.StyledText;
import org.eclipse.swt.dnd.DND;
import org.eclipse.swt.dnd.DragSource;
import org.eclipse.swt.dnd.DragSourceAdapter;
import org.eclipse.swt.dnd.DragSourceEvent;
import org.eclipse.swt.dnd.DragSourceListener;
import org.eclipse.swt.dnd.Transfer;
import org.eclipse.swt.events.FocusEvent;
import org.eclipse.swt.events.FocusListener;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.ui.forms.IManagedForm;
import org.eclipse.ui.forms.widgets.TableWrapData;
import org.eclipse.ui.forms.widgets.TableWrapLayout;

import eu.etaxonomy.cdm.model.taxon.Taxon;
import eu.etaxonomy.taxeditor.ITaxEditorConstants;
import eu.etaxonomy.taxeditor.TaxEditorPlugin;
import eu.etaxonomy.taxeditor.actions.WidgetTransfer;

/** 
 * Formats <code>GroupedComposite</code> with cosmetic and layout properties specific to the
 * Editor. This should be used to maintain a consistent look and feel for all Editor 
 * freetext area components, such as DescriptionElementComposite.
 * <p>
 * Requires an <code>IManagedForm</code>, whose <code>input</code> is set to the contents
 * of {@link #getData()} when the <code>GroupedComposite</code> gets focus, i.e. to
 * populate the property sheet with the data.
 * </p>
 * <p>
 * The <code>IManagedForm</code> is also required to have a <code>Taxon</code> in its 
 * own <code>getData()</code>.
 * </p> 
 * <p>
 * The <code>IManagedForm</code> can also used for drawing borders by calling the method 
 * <code>createBorderSupport()</code>.
 * </p>
 * @author p.ciardelli
 * @created 02.06.2008
 * @version 1.0
 */
abstract public class GroupedComposite extends Composite implements IHasPropertySource {	
	private static final Logger logger = Logger.getLogger(GroupedComposite.class);
	
	protected TextViewer textViewer;
	protected TableWrapLayout layout;
	private Label draggableLabel;
	protected IManagedForm managedForm;
	protected Taxon taxon;

	protected CompositeBorderDecorator borderDecorator;
	
	/**
	 * @param parent
	 * @param managedForm
	 */
	public GroupedComposite(Composite parent, IManagedForm managedForm) {
		super(parent, SWT.NONE);
		
		createContent();
		
		this.managedForm = managedForm;
		
		Object formData = managedForm.getForm().getBody().getData();
		Assert.isTrue(formData instanceof Taxon, 
				"Managed form must have a Taxon in its data field.");
		taxon = (Taxon) formData;
	}

	/**
	 * 
	 */
	protected void createLineWrapSupport() {
		new LineWrapSupport(getTextViewer(), managedForm);
	}
		
	protected void createContent() {
		setLayoutData(new TableWrapData(TableWrapData.FILL_GRAB));
		layout = new TableWrapLayout();
		layout.numColumns = 2;
		layout.leftMargin = 0;
		setLayout(layout);
		
		Color groupBackgroundColor = TaxEditorPlugin.getDefault().
				getColor(ITaxEditorConstants.GROUP_GRAY_BKG_COLOR);
		setBackground(groupBackgroundColor);		
		
		draggableLabel = new Label(this, SWT.NONE);
		draggableLabel.setBackground(groupBackgroundColor);
		draggableLabel.setLayoutData(new TableWrapData(TableWrapData.CENTER, TableWrapData.TOP));
		
		setDraggableControl(draggableLabel);
	}
		
	public Taxon getTaxon() {
		return taxon;
	}
	
	/**
	 * @return
	 */
	public TextViewer getTextViewer() {
		return textViewer;
	}

	/**
	 * @param textViewer
	 */
	public void setTextViewer(TextViewer textViewer) {
		this.textViewer = textViewer;
	}

	public void setText(String text) {
		Assert.isNotNull(getTextViewer(), 
				"Cannot set text for a TextViewer that has not yet been initialized.");
		Assert.isNotNull(getTextViewer().getDocument(), 
				"Cannot set text for a TextViewer whose Document has not yet been initialized.");
		getTextViewer().getDocument().set(text);
	}
	
	/**
	 * @param icon
	 */
	public void setIcon(Image icon) {
		draggableLabel.setImage(icon);
	}

	/**
	 * @return
	 */
	public Label getDraggableLabel() {
		return draggableLabel;
	}
	
	/**
	 * @param indent
	 */
	public void setIndent(int indent) {
		layout.leftMargin = indent;
		this.setLayout(layout);
	}	
	
	/* (non-Javadoc)
	 * @see org.eclipse.swt.widgets.Composite#setFocus()
	 */
	public boolean setFocus() {
		setSelection();
		return super.setFocus();
	}
	
	/**
	 * 
	 */
	protected void setSelection() {
		managedForm.setInput(this);
	}
	
	/* (non-Javadoc)
	 * @see org.eclipse.swt.widgets.Control#setFont(org.eclipse.swt.graphics.Font)
	 */
	public void setFont(Font font) {
		getTextViewer().getTextWidget().setFont(font);
	}

	FocusListener focusListener;
	
	/**
	 * If <code>textViewer</code> has already been set, it will show a 
	 * <code>prompt</code> along the lines of "Click here to start entering data" 
	 * when empty.
	 * 
	 * @param prompt
	 */
	public void createEmptyViewerPrompt(final String prompt) {
		Assert.isNotNull(getTextViewer());
//		new EmptyTextViewerPrompt(getTextViewer(), prompt);
		 
		final StyledText textControl = getTextViewer().getTextWidget();
		final IDocument document = getTextViewer().getDocument();
		final Font promptFont = TaxEditorPlugin.getDefault()
				.getFont(ITaxEditorConstants.DEFAULT_PROMPT_FONT);
		setFocusListener(new FocusListener() {

			@Override
			public void focusGained(FocusEvent e) {
				if (document.get().equals(prompt)) {
					textControl.setFont(getViewerFont());
					document.set("");
				}
			}

			@Override
			public void focusLost(FocusEvent e) {
				if (document.getLength() == 0) {
					initEmptyText();
				} else {
//					managedForm.reflow(true);
				}
			}
			
		});
		textControl.addFocusListener(getFocusListener());
		
		if (document.getLength() == 0) {
			textControl.setFont(promptFont);
			document.set(prompt);
		}
	}
	
	abstract protected Font getViewerFont();

	protected void initEmptyText() {
		getTextViewer().getTextWidget().setFont(
				TaxEditorPlugin.getDefault()
					.getFont(ITaxEditorConstants.DEFAULT_PROMPT_FONT));
		
		getTextViewer().getDocument().set(getEmptyTextPrompt());
	}

	protected String getEmptyTextPrompt() {
		return "Click to edit";
	}

	private void setFocusListener(FocusListener focusListener) {
		this.focusListener = focusListener;
	}

	private FocusListener getFocusListener() {
		return focusListener;
	}
	
//	abstract protected Font getNameCompositeFont();

	/* (non-Javadoc)
	 * @see org.eclipse.swt.widgets.Control#getMenu()
	 */
	public Menu getMenu () {
		
		Menu menu = draggableLabel.getMenu();
		if (menu == null) {
			menu = new Menu(draggableLabel);
			draggableLabel.setMenu(menu);
		}
		return menu;
	}
	
	/**
	 * 
	 */
	public void createBorderSupport() {
//		borderDecorator = new CompositeBorderDecorator(
//				this, managedForm);
		
		if (getTextViewer() == null) {
			logger.warn("Could not create border support - getTextViewer() returned null.");
		} else {
			borderDecorator = new CompositeBorderDecorator(
					getTextViewer().getTextWidget(), managedForm);		
//			borderDecorator.setBorderedComposite(getTextViewer().getTextWidget());
			borderDecorator.setLoseFocus(false);
			getTextViewer().getTextWidget().addFocusListener(borderDecorator);
		}
	}
	
	public void drawBorder() {
		borderDecorator.paintBorder();
	}

	/**
	 * @param isDirty
	 */
	protected void setDirty(boolean isDirty) {
		managedForm.dirtyStateChanged();
	}
	
	/* (non-Javadoc)
	 * @see eu.etaxonomy.taxeditor.editor.name.GroupedComposite#dispose()
	 */
	public void dispose () {
		
		logger.warn("disposing");
		
		// Nullify decorator to stop its listener to listening for
		//	composite focus
//		borderDecorator.dispose();
		
		if (dragSource != null) {
			dragSource.removeDragListener(dragSourceListener);
		}
		
		if (getFocusListener() != null && getTextViewer() != null 
				&& getTextViewer().getTextWidget() != null) {
			getTextViewer().getTextWidget().removeFocusListener(getFocusListener());
			getTextViewer().getTextWidget().removeFocusListener(borderDecorator);
			getTextViewer().getTextWidget().dispose();
		}

		super.dispose();
	}
	
	/**
	 * Returns a <code>ContextMenu</code> opened by right-clicking on
	 * <code>draggableLabel</code>.
	 *  
	 * @return
	 */
	protected ContextMenu createContextMenu() {
		ContextMenu contextMenu = new ContextMenu(getDraggableLabel());
		if (getTextViewer() != null) {
			getTextViewer().getTextWidget().setMenu(contextMenu.getMenu());
		}
		return contextMenu;
	}
	
	public void unpaintBorder() {
		if (borderDecorator != null) {
			borderDecorator.unpaintBorder();
		}
	}
	
	private Control draggableControl;
	private DragSource dragSource;
	
	protected void setDraggableControl(Control control) {
		draggableControl = control;
	}

	public void setDraggable(boolean draggable) {

		if (draggable) {

			if (dragSource != null) {
				// Already initialized
				return;
			}

			if (draggableControl == null) {
				throw new NullPointerException(
						"Draggable control must be set to add draggability");
			}
			
			Transfer[] types = new Transfer[] { WidgetTransfer.getInstance() };			
			int operations = DND.DROP_MOVE | DND.DROP_COPY | DND.DROP_LINK;

			dragSource = new DragSource(draggableControl, operations);
			dragSource.setTransfer(types);
			dragSource.addDragListener(dragSourceListener);

		} else {
			dragSource = null;
		}
	}

	/**
	 * Drag listener which passes the Composite as the data in a drag event.
	 */
	DragSourceListener dragSourceListener = new DragSourceAdapter() {

		public void dragStart(DragSourceEvent event) {
			GroupedComposite.this.setFocus();
			event.doit = true;
		}
		
		public void dragSetData(DragSourceEvent event) {
			WidgetTransfer.getInstance().setWidget(GroupedComposite.this);
		}
		
		public void dragFinished(DragSourceEvent event){
//			WidgetTransfer.getInstance().setWidget(null);
			dragSource = null;
//			((DragSource) event.widget).dispose();
		}
	};
} 