/**
* Copyright (C) 2007 EDIT
* European Distributed Institute of Taxonomy 
* http://www.e-taxonomy.eu
* 
* The contents of this file are subject to the Mozilla Public License Version 1.1
* See LICENSE.TXT at the top of this package for the full license terms.
*/

package eu.etaxonomy.taxeditor.model;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedSet;
import java.util.UUID;

import org.apache.log4j.Logger;
import org.eclipse.core.databinding.observable.Realm;
import org.eclipse.core.databinding.observable.set.IObservableSet;
import org.eclipse.core.databinding.observable.set.ISetChangeListener;
import org.eclipse.core.databinding.observable.set.SetChangeEvent;
import org.eclipse.core.databinding.observable.set.WritableSet;
import org.eclipse.jface.databinding.swt.SWTObservables;
import org.eclipse.swt.widgets.Display;

import eu.etaxonomy.cdm.api.application.CdmApplicationController;
import eu.etaxonomy.cdm.api.service.ITermService;
import eu.etaxonomy.cdm.model.common.Language;
import eu.etaxonomy.cdm.model.common.OrderedTermVocabulary;
import eu.etaxonomy.cdm.model.common.TermVocabulary;
import eu.etaxonomy.cdm.model.description.Feature;
import eu.etaxonomy.cdm.model.name.NameRelationshipType;
import eu.etaxonomy.cdm.model.name.NomenclaturalStatusType;
import eu.etaxonomy.cdm.model.name.Rank;
import eu.etaxonomy.cdm.model.reference.ReferenceBase;
import eu.etaxonomy.cdm.model.taxon.Taxon;
import eu.etaxonomy.cdm.model.taxon.TaxonRelationshipType;
import eu.etaxonomy.taxeditor.datasource.CdmTransactionController;
import eu.etaxonomy.taxeditor.editor.name.CdmParserController;
import eu.etaxonomy.taxeditor.navigation.RecentNamesView;

/**
 * @author p.ciardelli
 * @created 17.12.2008
 * @version 1.0
 */
public class CdmSessionDataRepository implements ICdmSessionDataRepository {
	private static final Logger logger = Logger
			.getLogger(CdmSessionDataRepository.class);

	private static ICdmSessionDataRepository repository;
	
	/**
	 * This map of taxa to their taxonomic children is used by
	 * NameTreeContentProvider's getChildren method.
	 * 
	 * key = taxon, value = key's child taxa Note that a map can have a key ==
	 * null, i.e. no parent taxon
	 */
	private HashMap<Taxon, Set<Taxon>> taxonomicChildrenMap;
	
	private CdmApplicationController applicationController;

	private IObservableSet observableTaxonSet;

	private List<ICdmTaxonSetListener> listeners;

	private TermVocabulary<Feature> features;

	private SortedSet<Rank> ranks;

	private TermVocabulary<NomenclaturalStatusType> nomStatus;

	private SortedSet<NameRelationshipType> nameRelationshipTypes;

	private OrderedTermVocabulary<TaxonRelationshipType> taxonRelationshipTypes;

	public static ICdmSessionDataRepository getDefault() {
		
		if (repository == null) {
			repository = new CdmSessionDataRepository();
		}
		
		
		return repository;
	}
	
	private CdmSessionDataRepository() {
		getObservableTaxa().addSetChangeListener(new observableTaxaListener());
	}
	
	@Override
	public void clearAllData() {
		clearTaxonData();
		clearNonTaxonData();
	}

	@Override
	public void clearTaxonData() {
		removeAllTaxa();
		
		// Clear "recent names" list
		RecentNamesView.clearRecentNames();
	}

	@Override
	public void clearNonTaxonData() {
		features = null;
		ranks = null;
		nomStatus = null;
		taxonRelationshipTypes = null;
		nameRelationshipTypes = null;
		
		// Parser is also transaction-, and therefore session-, dependent
		CdmParserController.clearNonViralNameParser();
	}

	@Override
	public void setApplicationController(CdmApplicationController applicationController) {
		this.applicationController = applicationController;
	}

	private Map<Taxon, Set<Taxon>> getTaxonomicChildrenMap() {
		if (taxonomicChildrenMap == null) {
			taxonomicChildrenMap = new HashMap<Taxon, Set<Taxon>>();
		}
		return taxonomicChildrenMap;
	}
	
	private CdmApplicationController getApplicationController() {
		// Make sure the app controller has been set
		if (applicationController == null) {
			throw new IllegalStateException("CdmApplicationController not yet set.");
		}
		return applicationController;
	}

	@Override
	public ReferenceBase getDefaultSec() {
		// TODO why is this returning null? and of course, replace w the real deal
		return getApplicationController().getReferenceService().getReferenceByUuid(
				UUID.fromString("f3593c18-a8d2-4e51-bdad-0befbf8fb2d1"));
	}

	@Override
	public Set<Taxon> getRootTaxa() {
		return getTaxonomicChildren(null);
	}
	
	@Override
	public Set<Taxon> getTaxonomicChildren(Taxon parentTaxon) {
				
		// Check whether the map contains children for parentTaxon
		if (!getTaxonomicChildrenMap().containsKey(parentTaxon)) {
			
			// Make a new child taxa set to ensure deleted taxa
			//	don't re-appear as NULL children
			Set<Taxon> childTaxa = new HashSet<Taxon>();
				
			// Root taxa have parentTaxon = NULL
			if (parentTaxon == null) {
				
				// Get root taxa from data source
				boolean onlyWithChildren = false;
				boolean withMisapplications = true;
				childTaxa.addAll(getApplicationController().getTaxonService().getRootTaxa(
					getDefaultSec(), onlyWithChildren, withMisapplications));
				
			} else {
				
				// Get children from taxon
				childTaxa.addAll(parentTaxon.getTaxonomicChildren());
			}
			
			// Iterate through all child taxa, throwing out NULLs and saving the rest to internal collections
			Set<Taxon> nullTaxa = new HashSet<Taxon>();
			for (Taxon taxon : childTaxa) {
				
				// Some imports resulted in NULL child taxa
				if (taxon == null) {
					if (parentTaxon == null) {
						logger.warn("NULL taxon in root taxa.");
					} else {
						logger.warn("Taxon " + parentTaxon.toString() + " has a NULL child taxon.");
					}
					nullTaxa.add(taxon);
				} else {
					
					// Add the taxon to internal collections
					addTaxonIntern(taxon);
				}
			}
			childTaxa.removeAll(nullTaxa);
							
			// Put the new set in the parent-children map
			getTaxonomicChildrenMap().put(parentTaxon, childTaxa);
		}
		return getTaxonomicChildrenMap().get(parentTaxon);
	}
	
	private void addTaxonIntern(Taxon taxon) {
				
		// Add it to the transaction
		CdmTransactionController.addTaxonToTransaction(taxon);

		// Remove then re-add session taxon
		getObservableTaxa().add(taxon);
	}
	
	@Override
	public void addTaxon(Taxon taxon) {
		
		// Add taxon to its parent's child map
		Taxon parentTaxon = taxon.getTaxonomicParent();
		getTaxonomicChildren(parentTaxon).add(taxon);
		
		// Add taxon to internal collections
		addTaxonIntern(taxon);
	}

	@Override
	public IObservableSet getObservableTaxa() {
		if (observableTaxonSet == null) {
			Realm realm = SWTObservables.getRealm(Display.getDefault());
			observableTaxonSet = new WritableSet(realm);
		}
		return observableTaxonSet;
	}

	@Override
	public boolean deleteTaxon(Taxon taxon) {
		
		// Remove taxon from repository collections
		removeTaxon(taxon);
		
		// If taxon has been saved before, delete it from the data source
		if (isInPersistentDataSource(taxon)) {
			getApplicationController().getTaxonService().removeTaxon(taxon);
		}
		
		// If parent taxon has been saved before, save it to reflect loss of child
		Taxon parentTaxon = taxon.getTaxonomicParent();
		if (parentTaxon != null && isInPersistentDataSource(parentTaxon)) {
			getApplicationController().getTaxonService().saveTaxon(parentTaxon);
		}
		
		return true;
	}
	
	private boolean isInPersistentDataSource(Taxon taxon) {
		UUID taxonUuid = taxon.getUuid();
		if (getApplicationController().getTaxonService()
				.getTaxonByUuid(taxonUuid) != null) {
			return true;
		} else {
			return false;
		}
	}

	@Override
	public void removeTaxon(Taxon taxon) {

		// Remove from session taxa
		getObservableTaxa().remove(taxon);
		
		// Recursively remove all children, children's children, etc.
		clearTaxonomicChildren(taxon);

		// Remove from parent and from parent's child map
		Taxon parentTaxon = taxon.getTaxonomicParent();
		if (parentTaxon != null) {
			removeTaxonomicChild(taxon, parentTaxon);
		}
		
		// Remove from recent names list
		RecentNamesView.removeRecentName(taxon);
	}
	
	private void removeTaxonomicChild(Taxon taxon, Taxon parentTaxon) {
		parentTaxon.removeTaxonomicChild(taxon);
		getTaxonomicChildren(parentTaxon).remove(taxon);
	}

	private void clearTaxonomicChildren(Taxon taxon) {
		Set<Taxon> children = getTaxonomicChildren(taxon);
		if (children != null) {
			for (Taxon child : children) {
				clearTaxonomicChildren(child);
			}
			getTaxonomicChildrenMap().remove(taxon);
			getObservableTaxa().removeAll(children);
		}
	}

	@Override
	public Collection<Taxon> getAllTaxa() {
		return getObservableTaxa();
	}
	
	@Override
	public void removeAllTaxa() {
		taxonomicChildrenMap.clear();
		observableTaxonSet.clear();
	}


	@Override
	public boolean saveTaxon(Taxon taxon) {
		
		// If this is not in the list of observable taxa, add
		if (!isInRepository(taxon)) {
			addTaxon(taxon);
		}
		
		// Save taxon to CDM layer
		CdmUtil.getTaxonService().saveTaxon(taxon);
		
		// Notify taxon listeners in case name has been updated - 
		//	updates editor view tab and recent names
		taxon.firePropertyChange("name", null, null);
		
		return true;
	}
	
	private boolean isInRepository(Taxon taxon) {
		return observableTaxonSet.contains(taxon);
	}

	@Override
	public void setTaxonomicParent(Taxon taxon, Taxon newParentTaxon) {
		
		// Get old taxonomic parent
		Taxon oldParentTaxon = taxon.getTaxonomicParent();
		
		// Set new taxonomic parent
		taxon.setTaxonomicParent(newParentTaxon, null, null);
		
		// Update child taxa collection		
		getTaxonomicChildren(oldParentTaxon).remove(taxon);
		getTaxonomicChildren(newParentTaxon).add(taxon);
		
		// Notify listeners that taxon has moved
		if (listeners != null) {
			for (ICdmTaxonSetListener listener : listeners) {
				listener.taxonMoved(taxon, newParentTaxon);
			}
		}
	}
	
	@Override
	public TermVocabulary<Feature> getFeatures() {
		if (features == null) {
			features = getApplicationController().getDescriptionService().
					getDefaultFeatureVocabulary();
		}
		return features;
	}
	
	@Override
	public SortedSet<Rank> getRanks() {
		if (ranks == null) {
			
			OrderedTermVocabulary<Rank> rankVocabulary = 
					getApplicationController().getNameService().getRankVocabulary();
			ranks = rankVocabulary.getOrderedTerms(null);
		}
		return ranks;
	}

	@Override
	public TermVocabulary<NomenclaturalStatusType> getNomStatus() {
		if (nomStatus == null) {
			nomStatus = getApplicationController().getNameService()
					.getStatusTypeVocabulary();
		}
		return nomStatus;
	}
	
	@Override
	public SortedSet<NameRelationshipType> getNameRelationshipTypes() {
		if (nameRelationshipTypes == null) {
			TermVocabulary<NameRelationshipType> nameRelationshipTypesUnsorted = 
					getApplicationController().getNameService()
						.getNameRelationshipTypeVocabulary();
			
			// Add all terms manually to transaction
			ITermService termService = getApplicationController().getTermService();
			for (NameRelationshipType type : nameRelationshipTypesUnsorted) {
				termService.saveTerm(type);
			}
			
			// Get sorted list of types
			nameRelationshipTypes = nameRelationshipTypesUnsorted.
					getTermsOrderedByLabels(Language.DEFAULT());
		}
		return nameRelationshipTypes;
	}
	
	@Override
	public void addTaxonSetListener(ICdmTaxonSetListener listener) {
		
		if (listeners == null) { 
			listeners = new ArrayList<ICdmTaxonSetListener>();
		}
		listeners.add(listener);
	}
	
	class observableTaxaListener implements ISetChangeListener {
		@Override
		public void handleSetChange(SetChangeEvent event) {
			if (event.diff != null) {
				Set<Taxon> additions = event.diff.getAdditions();
				Set<Taxon> removals = event.diff.getRemovals();
				
				if (listeners != null) {
					for (ICdmTaxonSetListener listener : listeners) {
						
						if (additions.size() > 0) {
							listener.taxaAdded(additions);
						}
						
						if (removals.size() > 0) {
							listener.taxaRemoved(removals);
						}
					}
				}
			}
		}		
	}

	@Override
	public OrderedTermVocabulary<TaxonRelationshipType> getTaxonRelationshipTypes() {
		if (taxonRelationshipTypes == null) {
			taxonRelationshipTypes = getApplicationController().getTaxonService().
					getTaxonRelationshipTypeVocabulary();
		}
		return taxonRelationshipTypes;
	}
}
